==> liferea_1.4.4.orig.tar.gz: liferea-1.4.4/src/comments.c <==
/**
 * @file comments.c comment feed handling
 * 
 * Copyright (C) 2007 Lars Lindner <lars.lindner@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include "comments.h"
#include "common.h"
#include "db.h"
#include "debug.h"
#include "feed.h"
#include "itemview.h"
#include "metadata.h"
#include "update.h"

/* Comment feeds in Liferea are simple flat lists of items attached
   to a single item. Each item that has a comment feed URL in its 
   metadata list gets its comment feed updated as soon as the user
   triggers rendering of the item in 3 pane mode.
   
   Although rendered differently items and comment items are handled
   in the same way. */

static GHashTable	*commentFeeds = NULL;

typedef struct commentFeed 
{
	gulong		itemId;			/**< parent item id */
	gchar		*id;			/**< id of the items comments feed (or NULL) */
	gchar		*error;			/**< description of error if comments download failed (or NULL)*/

	struct updateJob *updateJob;		/**< update job structure used when downloading comments */
	updateStatePtr	updateState;		/**< update states (etag, last modified, cookies, last polling times...) used when downloading comments */
} *commentFeedPtr;

static void
comment_feed_free (commentFeedPtr commentFeed)
{
	if (commentFeed->updateJob)
		update_job_cancel_by_owner (commentFeed);
	if (commentFeed->updateState)
		update_state_free (commentFeed->updateState);
	
	g_free (commentFeed->error);
	g_free (commentFeed->id);
	g_free (commentFeed);
}

static void
comment_feed_free_cb (gpointer key, gpointer value, gpointer user_data)
{
	comment_feed_free (value);
}

void
comments_deinit (void)
{
	if (commentFeeds) {
		g_hash_table_foreach (commentFeeds, comment_feed_free_cb, NULL);
		g_hash_table_destroy (commentFeeds);
		commentFeeds = NULL;
	}
}

/**
 * Hash lookup to find comment feeds with the given id.
 * Returns the comment feed (or NULL).
 */
static commentFeedPtr
comment_feed_from_id (const gchar *id)
{
	if (!commentFeeds)
		return NULL;
		
	return (commentFeedPtr) g_hash_table_lookup (commentFeeds, id);
}

static void
comments_process_update_result (const struct updateResult * const result, gpointer user_data, updateFlags flags) 
{
	feedParserCtxtPtr	ctxt;
	commentFeedPtr		commentFeed = (commentFeedPtr)user_data;
	itemPtr			item;
	nodePtr			node;

	debug_enter ("comments_process_update_result");
	
	item = item_load (commentFeed->itemId);
	g_return_if_fail (item != NULL);
	
	/* note this is to update the feed URL on permanent redirects */
	if (!strcmp (result->source, metadata_list_get (item->metadata, "commentFeedUri"))) {
	
		debug2 (DEBUG_UPDATE, "updating comment feed URL from \"%s\" to \"%s\"", 
		                      metadata_list_get (item->metadata, "commentFeedUri"), 
				      result->source);
				     
		metadata_list_set (&(item->metadata), "commentFeedUri", result->source);
	}
	
	if (401 == result->httpstatus) { /* unauthorized */
		commentFeed->error = g_strdup (_("Authorization Error"));
	} else if (410 == result->httpstatus) { /* gone */
		// FIXME: how to prevent further updates?
	} else if (304 == result->httpstatus) {
		debug1(DEBUG_UPDATE, "comment feed \"%s\" did not change", result->source);
	} else if (result->data) {
		debug1(DEBUG_UPDATE, "received update result for comment feed \"%s\"", result->source);

		/* parse the new downloaded feed into fake node, subscription and feed */
		node = node_new ();
		ctxt = feed_create_parser_ctxt ();
		ctxt->subscription = subscription_new (result->source, NULL, NULL);
		ctxt->feed = feed_new ();
		node_set_type (node, feed_get_node_type ());
		node_set_data (node, ctxt->feed);		
		node_set_subscription (node, ctxt->subscription);
		ctxt->data = result->data;
		ctxt->dataLength = result->size;
		feed_parse (ctxt);

		if (ctxt->failed) {
			debug0 (DEBUG_UPDATE, "parsing comment feed failed!");
		} else {
			itemSetPtr	comments;
			GList		*iter;
			
			/* before merging mark all downloaded items as comments */
			iter = ctxt->items;
			while (iter) {
				((itemPtr) iter->data)->isComment = TRUE;
				iter = g_list_next (iter);
			}
			
			debug1 (DEBUG_UPDATE, "parsing comment feed successful (%d comments downloaded)", g_list_length(ctxt->items));		
			comments = db_itemset_load (commentFeed->id);
			itemset_merge_items (comments, ctxt->items);
			itemset_free (comments);
		}
				
		node_free (ctxt->subscription->node);
		feed_free_parser_ctxt (ctxt);
	}
	
	/* update error message */
	g_free (commentFeed->error);
	commentFeed->error = NULL;
	
	if (!(result->httpstatus >= 200) && (result->httpstatus < 400)) {
		const gchar * tmp;
		
		/* first specific codes (guarantees tmp to be set) */
		tmp = common_http_error_to_str (result->httpstatus);

		/* second netio errors */
		if (common_netio_error_to_str (result->returncode))
			tmp = common_netio_error_to_str (result->returncode);
			
		commentFeed->error = g_strdup (tmp);
	}	

	/* clean up... */
	commentFeed->updateJob = NULL;

	/* rerender item with new comments */
	itemview_update_item (item); 
	itemview_update ();
	
	item_unload (item);
		
	debug_exit ("comments_process_update_result");
}

void
comments_refresh (itemPtr item) 
{ 
	commentFeedPtr		commentFeed = NULL;
	updateRequestPtr	request;
	const gchar		*url;
	
	if (!update_is_online ())
		return;
	
	url = metadata_list_get (item->metadata, "commentFeedUri");
	if (url) {
		debug2 (DEBUG_UPDATE, "Updating comments for item \"%s\" (comment URL: %s)", item->title, url);

		// FIXME: restore update state from DB?		
		
		if (item->commentFeedId)
			commentFeed = comment_feed_from_id (item->commentFeedId);
		else
			item->commentFeedId = node_new_id ();		
			
		if (!commentFeed) {			
			commentFeed = g_new0 (struct commentFeed, 1);
			commentFeed->id = g_strdup (item->commentFeedId);
			commentFeed->itemId = item->id;
			commentFeed->updateState = update_state_new ();

			if (!commentFeeds)
				commentFeeds = g_hash_table_new (g_str_hash, g_str_equal);
			g_hash_table_insert (commentFeeds, commentFeed->id, commentFeed);
		}

		request = update_request_new ();
		request->options = g_new0 (struct updateOptions, 1);	// FIXME: use copy of parent subscription options
		request->source = g_strdup (url);
		commentFeed->updateJob = update_execute_request (commentFeed, request, comments_process_update_result, commentFeed, FEED_REQ_PRIORITY_HIGH);

		/* Item view refresh to change link from "Update" to "Updating..." */
		itemview_update_item (item); 
		itemview_update ();
	}
}

void
comments_to_xml (xmlNodePtr parentNode,
                 const gchar *id)
{
	xmlNodePtr	commentsNode;
	commentFeedPtr	commentFeed;
	itemSetPtr	itemSet;
	GList		*iter;

	commentsNode = xmlNewChild (parentNode, NULL, "comments", NULL);
	
	commentFeed = comment_feed_from_id (id);
	g_return_if_fail (commentFeed != NULL);
	
	itemSet = db_itemset_load (id);
	g_return_if_fail (itemSet != NULL);

	iter = itemSet->ids;
	while (iter) 
	{
		itemPtr comment = item_load (GPOINTER_TO_UINT (iter->data));
		item_to_xml (comment, commentsNode);
		item_unload (comment);
		iter = g_list_next (iter);
	}

	xmlNewTextChild (commentsNode, NULL, "updateState", 
	                 (commentFeed->updateJob)?"updating":"ok");
		
	if (commentFeed->error)
		xmlNewTextChild (commentsNode, NULL, "updateError", commentFeed->error);
			
	itemset_free (itemSet);
}

void
comments_remove (const gchar *id)
{
	commentFeedPtr commentFeed;

	db_itemset_remove_all (id);
	
	commentFeed = comment_feed_from_id (id);
	g_return_if_fail (commentFeed != NULL);
	comment_feed_free (commentFeed);
}

==> liferea-1.4.12.tar.gz: liferea-1.4.12/src/comments.c <==
/**
 * @file comments.c comment feed handling
 * 
 * Copyright (C) 2007 Lars Lindner <lars.lindner@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include "comments.h"
#include "common.h"
#include "db.h"
#include "debug.h"
#include "feed.h"
#include "itemview.h"
#include "metadata.h"
#include "update.h"

/* Comment feeds in Liferea are simple flat lists of items attached
   to a single item. Each item that has a comment feed URL in its 
   metadata list gets its comment feed updated as soon as the user
   triggers rendering of the item in 3 pane mode.
   
   Although rendered differently items and comment items are handled
   in the same way. */

static GHashTable	*commentFeeds = NULL;

typedef struct commentFeed 
{
	gulong		itemId;			/**< parent item id */
	gchar		*id;			/**< id of the items comments feed (or NULL) */
	gchar		*error;			/**< description of error if comments download failed (or NULL)*/

	struct updateJob *updateJob;		/**< update job structure used when downloading comments */
	updateStatePtr	updateState;		/**< update states (etag, last modified, cookies, last polling times...) used when downloading comments */
} *commentFeedPtr;

static void
comment_feed_free (commentFeedPtr commentFeed)
{
	if (commentFeed->updateJob)
		update_job_cancel_by_owner (commentFeed);
	if (commentFeed->updateState)
		update_state_free (commentFeed->updateState);
	
	g_free (commentFeed->error);
	g_free (commentFeed->id);
	g_free (commentFeed);
}

static void
comment_feed_free_cb (gpointer key, gpointer value, gpointer user_data)
{
	comment_feed_free (value);
}

void
comments_deinit (void)
{
	if (commentFeeds) {
		g_hash_table_foreach (commentFeeds, comment_feed_free_cb, NULL);
		g_hash_table_destroy (commentFeeds);
		commentFeeds = NULL;
	}
}

/**
 * Hash lookup to find comment feeds with the given id.
 * Returns the comment feed (or NULL).
 */
static commentFeedPtr
comment_feed_from_id (const gchar *id)
{
	if (!commentFeeds)
		return NULL;
		
	return (commentFeedPtr) g_hash_table_lookup (commentFeeds, id);
}

static void
comments_process_update_result (const struct updateResult * const result, gpointer user_data, updateFlags flags) 
{
	feedParserCtxtPtr	ctxt;
	commentFeedPtr		commentFeed = (commentFeedPtr)user_data;
	itemPtr			item;
	nodePtr			node;

	debug_enter ("comments_process_update_result");
	
	item = item_load (commentFeed->itemId);
	g_return_if_fail (item != NULL);
	
	/* note this is to update the feed URL on permanent redirects */
	if (!strcmp (result->source, metadata_list_get (item->metadata, "commentFeedUri"))) {
	
		debug2 (DEBUG_UPDATE, "updating comment feed URL from \"%s\" to \"%s\"", 
		                      metadata_list_get (item->metadata, "commentFeedUri"), 
				      result->source);
				     
		metadata_list_set (&(item->metadata), "commentFeedUri", result->source);
	}
	
	if (401 == result->httpstatus) { /* unauthorized */
		commentFeed->error = g_strdup (_("Authorization Error"));
	} else if (410 == result->httpstatus) { /* gone */
		// FIXME: how to prevent further updates?
	} else if (304 == result->httpstatus) {
		debug1(DEBUG_UPDATE, "comment feed \"%s\" did not change", result->source);
	} else if (result->data) {
		debug1(DEBUG_UPDATE, "received update result for comment feed \"%s\"", result->source);

		/* parse the new downloaded feed into fake node, subscription and feed */
		node = node_new ();
		ctxt = feed_create_parser_ctxt ();
		ctxt->subscription = subscription_new (result->source, NULL, NULL);
		ctxt->feed = feed_new ();
		node_set_type (node, feed_get_node_type ());
		node_set_data (node, ctxt->feed);		
		node_set_subscription (node, ctxt->subscription);
		ctxt->data = result->data;
		ctxt->dataLength = result->size;
		feed_parse (ctxt);

		if (ctxt->failed) {
			debug0 (DEBUG_UPDATE, "parsing comment feed failed!");
		} else {
			itemSetPtr	comments;
			GList		*iter;
			
			/* before merging mark all downloaded items as comments */
			iter = ctxt->items;
			while (iter) {
				itemPtr comment = (itemPtr) iter->data;
				comment->isComment = TRUE;
				comment->parentItemId = commentFeed->itemId;				
				iter = g_list_next (iter);
			}
			
			debug1 (DEBUG_UPDATE, "parsing comment feed successful (%d comments downloaded)", g_list_length(ctxt->items));		
			comments = db_itemset_load (commentFeed->id);
			itemset_merge_items (comments, ctxt->items, ctxt->feed->valid);
			itemset_free (comments);
			
			/* No comment feed truncating as comment items are automatically
			   dropped when the parent items are removed from cache. */
		}
				
		node_free (ctxt->subscription->node);
		feed_free_parser_ctxt (ctxt);
	}
	
	/* update error message */
	g_free (commentFeed->error);
	commentFeed->error = NULL;
	
	if (!(result->httpstatus >= 200) && (result->httpstatus < 400)) {
		const gchar * tmp;
		
		/* first specific codes (guarantees tmp to be set) */
		tmp = common_http_error_to_str (result->httpstatus);

		/* second netio errors */
		if (common_netio_error_to_str (result->returncode))
			tmp = common_netio_error_to_str (result->returncode);
			
		commentFeed->error = g_strdup (tmp);
	}	

	/* clean up... */
	commentFeed->updateJob = NULL;

	/* rerender item with new comments */
	itemview_update_item (item); 
	itemview_update ();
	
	item_unload (item);
		
	debug_exit ("comments_process_update_result");
}

void
comments_refresh (itemPtr item) 
{ 
	commentFeedPtr		commentFeed = NULL;
	updateRequestPtr	request;
	const gchar		*url;
	
	if (!update_is_online ())
		return;
	
	url = metadata_list_get (item->metadata, "commentFeedUri");
	if (url) {
		debug2 (DEBUG_UPDATE, "Updating comments for item \"%s\" (comment URL: %s)", item->title, url);

		// FIXME: restore update state from DB?		
		
		if (item->commentFeedId)
			commentFeed = comment_feed_from_id (item->commentFeedId);
		else
			item->commentFeedId = node_new_id ();		
			
		if (!commentFeed) {			
			commentFeed = g_new0 (struct commentFeed, 1);
			commentFeed->id = g_strdup (item->commentFeedId);
			commentFeed->itemId = item->id;
			commentFeed->updateState = update_state_new ();

			if (!commentFeeds)
				commentFeeds = g_hash_table_new (g_str_hash, g_str_equal);
			g_hash_table_insert (commentFeeds, commentFeed->id, commentFeed);
		}

		request = update_request_new ();
		request->options = g_new0 (struct updateOptions, 1);	// FIXME: use copy of parent subscription options
		request->source = g_strdup (url);
		commentFeed->updateJob = update_execute_request (commentFeed, request, comments_process_update_result, commentFeed, FEED_REQ_PRIORITY_HIGH);

		/* Item view refresh to change link from "Update" to "Updating..." */
		itemview_update_item (item); 
		itemview_update ();
	}
}

void
comments_to_xml (xmlNodePtr parentNode, const gchar *id)
{
	xmlNodePtr	commentsNode;
	commentFeedPtr	commentFeed;
	itemSetPtr	itemSet;
	GList		*iter;
	
	commentFeed = comment_feed_from_id (id);
	if (!commentFeed)
		return;

	commentsNode = xmlNewChild (parentNode, NULL, "comments", NULL);
	
	itemSet = db_itemset_load (id);
	g_return_if_fail (itemSet != NULL);

	iter = itemSet->ids;
	while (iter) 
	{
		itemPtr comment = item_load (GPOINTER_TO_UINT (iter->data));
		item_to_xml (comment, commentsNode);
		item_unload (comment);
		iter = g_list_next (iter);
	}

	xmlNewTextChild (commentsNode, NULL, "updateState", 
	                 (commentFeed->updateJob)?"updating":"ok");
		
	if (commentFeed->error)
		xmlNewTextChild (commentsNode, NULL, "updateError", commentFeed->error);
			
	itemset_free (itemSet);
}

void
comments_remove (const gchar *id)
{
	commentFeedPtr commentFeed;

	db_itemset_remove_all (id);
	
	commentFeed = comment_feed_from_id (id);
	g_return_if_fail (commentFeed != NULL);
	comment_feed_free (commentFeed);
}
